;;;
;;; Universal Monitor NS32000
;;;   Copyright (C) 2023 Haruo Asano
;;;

	CPU	NS32016
	SUPMODE	ON

	INCLUDE	"config.inc"
	INCLUDE	"../common.inc"

;;;
;;; ROM area
;;;
	
	ORG	0000H

CSTART:
	LPRD	SP,STACK
	LPRD	CFG,CFG_V
	LPRD	SB,WORK_B
	LPRW	MOD,MODDSC

	BSR	INIT

	MOVD	RAM_B,R0
	MOVD	R0,@DSADDR
	MOVD	R0,@SADDR
	MOVD	R0,@GADDR
	MOVB	'S',@HEXMOD

	MOVD	OPNMSG,R1
	BSR	STROUT

WSTART:
	MOVD	PROMPT,R1
	BSR	STROUT
	BSR	GETLIN
	MOVD	INBUF,R1
	BSR	SKIPSP
	MOVB	(R1),R0
	BSR	UPPER
	CMPB	00H,R0
	BEQ	WSTART

	CMPB	'D',R0
	BEQ	DUMP
	CMPB	'G',R0
	BEQ	GO
	CMPB	'S',R0
	BEQ	SETM

	CMPB	'L',R0
	BEQ	LOADH

ERR:
	MOVD	ERRMSG,R1
	BSR	STROUT
	BR	WSTART

;;; Dump memory

DUMP:
	ADDQD	1,R1
	BSR	SKIPSP
	BSR	RDHEX		; 1st arg.
	CMPB	0,R2
	BNE	DP0
	;; No arg.
	BSR	SKIPSP
	CMPB	00H,(R1)
	BNE	ERR
	MOVD	@DSADDR,R0
	ADDD	128,R0
	MOVD	R0,@DEADDR
	BR	DPM

	;; 1st arg. found
DP0:
	MOVD	R3,@DSADDR
	BSR	SKIPSP
	CMPB	',',(R1)
	BEQ	DP1
	CMPB	00H,(R1)
	BNE	ERR
	;; No 2nd arg.
	ADDD	128,R3
	MOVD	R3,@DEADDR
	BR	DPM

DP1:
	ADDQD	1,R1
	BSR	SKIPSP
	BSR	RDHEX
	BSR	SKIPSP
	CMPB	0,R2
	BEQ	ERR
	CMPB	00H,(R1)
	BNE	ERR
	ADDQD	1,R3
	MOVD	R3,@DEADDR

	;; DUMP main
DPM:
	MOVD	@DSADDR,R0
	ANDB	0F0H,R0
	MOVD	R0,R3
	XORB	R2,R2		; DSTATE
DPM0:
	BSR	DPL
	BSR	CONST
	CMPB	0,R0
	BNE	DPM1
	CMPB	2,R2		; DSTATE
	BHI	DPM0
	MOVD	@DEADDR,@DSADDR
	BR	WSTART
DPM1:
	MOVD	R3,@DSADDR
	BSR	CONIN
	BR	WSTART

	;; DUMP line
DPL:
	MOVD	R3,R0
	BSR	HEXOUT8
	MOVD	DSEP0,R1
	BSR	STROUT
	MOVD	INBUF,R4
	MOVB	16,R5
DPL0:
	BSR	DPB
	ACBB	-1,R5,DPL0

	MOVD	DSEP1,R1
	BSR	STROUT

	MOVD	INBUF,R4
	MOVB	16,R5
DPL1:
	MOVB	(R4),R0
	ADDQD	1,R4
	CMPB	' ',R0
	BHI	DPL2
	CMPB	7FH,R0
	BLS	DPL2
	BSR	CONOUT
	BR	DPL3
DPL2:
	MOVB	'.',R0
	BSR	CONOUT
DPL3:
	ACBB	-1,R5,DPL1
	BR	CRLF

	;; DUMP byte
DPB:
	MOVB	' ',R0
	BSR	CONOUT
	CMPB	0,R2		; DSTATE
	BNE	DPB2
	;; Dump state 0
	CMPD	@DSADDR,R3
	BEQ	DPB1
	;; Still 0 or 2
DPB0:
	MOVB	' ',R0
	MOVB	R0,(R4)
	BSR	CONOUT
	MOVB	' ',R0
	BSR	CONOUT
	ADDQD	1,R4
	ADDQD	1,R3
	RET
	;; Found start address
DPB1:
	MOVQB	1,R2		; DSTATE
DPB2:
	CMPB	1,R2		; DSTATE
	BNE	DPB0
	;; Dump state 1
	MOVB	(R3),R0
	MOVB	R0,(R4)
	ADDQD	1,R3
	ADDQD	1,R4
	BSR	HEXOUT2
	CMPD	@DEADDR,R3
	BEQ	DPB3
	RET
	;; Found end address
DPB3:
	MOVQB	2,R2		; DSTATE
	RET

;;; GO address

GO:
	ADDQD	1,R1
	BSR	SKIPSP
	BSR	RDHEX
	BSR	SKIPSP
	CMPB	00H,(R1)
	BNE	ERR
	CMPB	0,R2
	BEQ	G0
	MOVD	R3,@GADDR
G0:
	MOVD	@GADDR,R0
	JUMP	(R0)

;;; SET memory

SETM:
	ADDQD	1,R1
	BSR	SKIPSP
	BSR	RDHEX
	BSR	SKIPSP
	CMPB	00H,(R1)
	BNE	ERR
	MOVD	R3,R4
	CMPB	0,R2
	BNE	SM0
	MOVD	@SADDR,R4
SM0:
SM1:
	MOVD	R4,R0
	BSR	HEXOUT8
	MOVD	DSEP1,R1
	BSR	STROUT
	MOVB	(R4),R0
	BSR	HEXOUT2
	MOVB	' ',R0
	BSR	CONOUT

	BSR	GETLIN
	MOVD	INBUF,R1
	BSR	SKIPSP
	MOVB	(R1),R0
	CMPB	00H,R0
	BNE	SM2
	;; Empty (Increment address)
	ADDQD	1,R4
	MOVD	R4,@SADDR
	BR	SM1
SM2:
	CMPB	'-',R0
	BNE	SM3
	;; '-' (Decrement address)
	ADDQD	-1,R4
	MOVD	R4,@SADDR
	BR	SM1
SM3:
	CMPB	'.',R0
	BNE	SM4
	;; '.' (Quit)
	MOVD	R4,@SADDR
	BR	WSTART
SM4:
	BSR	RDHEX
	CMPB	0,R2
	BEQ	ERR
	MOVB	R3,(R4)
	ADDQD	1,R4
	MOVD	R4,@SADDR
	BR	SM1

;;; LOAD HEX file

LOADH:
	ADDQD	1,R1
	BSR	SKIPSP
	BSR	RDHEX
	BSR	SKIPSP
	CMPB	00H,(R1)
	BNE	ERR

	CMPB	0,R2
	BNE	LH0

	XORD	R3,R3		; Offset (Is this really needed?)
LH0:
	BSR	CONIN
	BSR	UPPER
	CMPB	'S',R0
	BEQ	LHS0
LH1:
	CMPB	':',R0
	BEQ	LHI0
	;; Skip to EOL
LH2:
	CMPB	CR,R0
	BEQ	LH0
	CMPB	LF,R0
	BEQ	LH0
LH3:
	BSR	CONIN
	BR	LH2

	;; Intel HEX
LHI0:
	BSR	HEXIN
	MOVB	R0,R4		; Length
	MOVB	R0,R5		; Checksum

	BSR	HEXIN
	XORD	R2,R2
	MOVB	R0,R2		; Address H
	ADDB	R0,R5		; Checksum

	BSR	HEXIN
	LSHD	8,R2
	MOVB	R0,R2		; Address L
	ADDB	R0,R5		; Checksum

	;; Add Offset
	ADDD	R3,R2

	BSR	HEXIN
	MOVB	R0,R6		; RECTYP
	ADDB	R0,R5		; Checksum

	CMPB	0,R4
	BEQ	LHI3
LHI1:
	BSR	HEXIN
	ADDB	R0,R5		; Checksum

	CMPB	0,R6		; RECTYP
	BNE	LHI2

	MOVB	R0,(R2)
	ADDQD	1,R2
LHI2:
	ACBB	-1,R4,LHI1
LHI3:
	BSR	HEXIN
	ADDB	R0,R5		; Checksum
	CMPB	0,R5
	BNE	LHIE		; Checksum error
	CMPB	0,R6		; RECTYP
	BEQ	LH3
	BR	WSTART
LHIE:
	MOVD	IHEMSG,R1
	BSR	STROUT
	BR	WSTART

	;; Motorola S record
LHS0:
	BSR	CONIN
	XORB	R7,R7		; Clear end flag
	MOVQB	2,R6		; 2 byte address
	CMPB	'1',R0
	BEQ	LHS01
	CMPB	'9',R0
	BEQ	LHS00
	MOVQB	3,R6		; 3 byte address
	CMPB	'2',R0
	BEQ	LHS01
	CMPB	'8',R0
	BEQ	LHS00
	MOVQB	4,R6		; 4 byte address
	CMPB	'3',R0
	BEQ	LHS01
	CMPB	'7',R0
	BEQ	LHS00
	BR	LH3		; Skip unsupported record type
LHS00:
	MOVQB	1,R7		; RECTYP (end mark)
LHS01:
	BSR	HEXIN
	MOVB	R0,R4		; Length
	MOVB	R0,R5		; Checksum

	XORD	R2,R2
LHS02:
	BSR	HEXIN		; address
	LSHD	8,R2
	MOVB	R0,R2
	ADDB	R0,R5		; Checksum

	ADDQB	-1,R4
	ACBB	-1,R6,LHS02

	;; Add offset
	ADDD	R3,R2

	ADDQB	-1,R4
	CMPB	0,R4
	BEQ	LHS3
LHS1:
	BSR	HEXIN
	ADDB	R0,R5		; Checksum

	CMPB	0,R7
	BNE	LHS2		; End mark

	MOVB	R0,(R2)
	ADDQD	1,R2
LHS2:
	ACBB	-1,R4,LHS1
LHS3:
	BSR	HEXIN
	ADDB	R0,R5		; Checksum
	CMPB	0FFH,R5
	BNE	LHSE		; Checksum error

	CMPB	0,R7
	BEQ	LH3
	BR	WSTART
LHSE:
	MOVD	SHEMSG,R1
	BSR	STROUT
	BR	WSTART
	
;;; Other support routines

STROUT:
	MOVB	(R1),R0
	ADDQD	1,R1
	CMPB	0,R0
	BEQ	STROE
	BSR	CONOUT
	BR	STROUT
STROE:
	RET

HEXOUT8:
	MOVD	R0,TOS
	ROTD	16,R0
	BSR	HEXOUT4
	MOVD	TOS,R0
HEXOUT4:
	MOVD	R0,TOS
	ROTW	8,R0
	BSR	HEXOUT2
	MOVD	TOS,R0
HEXOUT2:
	MOVD	R0,TOS
	ROTB	4,R0
	BSR	HEXOUT1
	MOVD	TOS,R0
HEXOUT1:
	ANDB	0FH,R0
	ADDB	'0',R0
	CMPB	'9'+1,R0
	BHI	CONOUT
	ADDB	'A'-'9'-1,R0
	BR	CONOUT

HEXIN:
	XORB	R0,R0
	BSR	HI0
	LSHB	4,R0
HI0:
	MOVB	R0,TOS
	BSR	CONIN
	BSR	UPPER
	CMPB	'0',R0
	BHI	HIR
	CMPB	'9'+1,R0
	BHI	HI1
	CMPB	'A',R0
	BHI	HIR
	CMPB	'F'+1,R0
	BLS	HIR
	SUBB	'A'-'9'-1,R0
HI1:
	SUBB	'0',R0
	ORB	TOS,R0
	RET
HIR:
	MOVB	TOS,R0
	RET

CRLF:
	MOVB	CR,R0
	BSR	CONOUT
	MOVB	LF,R0
	BR	CONOUT

GETLIN:
	MOVD	INBUF,R1
	XORB	R2,R2
GL0:
	BSR	CONIN
	CMPB	CR,R0
	BEQ	GLE
	CMPB	LF,R0
	BEQ	GLE
	CMPB	BS,R0
	BEQ	GLB
	CMPB	DEL,R0
	BEQ	GLB
	CMPB	' ',R0
	BHI	GL0
	CMPB	80H,R0
	BLS	GL0
	CMPB	BUFLEN-1,R2
	BLS	GL0		; Too long
	ADDQB	1,R2
	MOVB	R0,(R1)
	ADDQD	1,R1
	BSR	CONOUT
	BR	GL0
GLB:
	CMPB	0,R2
	BEQ	GL0
	ADDQD	-1,R1
	ADDQB	-1,R2
	MOVB	BS,R0
	BSR	CONOUT
	MOVB	' ',R0
	BSR	CONOUT
	MOVB	BS,R0
	BSR	CONOUT
	BR	GL0
GLE:
	BSR	CRLF
	MOVB	00H,(R1)
	RET

SKIPSP:
	CMPB	' ',(R1)
	ADDQD	1,R1
	BEQ	SKIPSP
	ADDQD	-1,R1
	RET

UPPER:
	CMPB	'a',R0
	BHI	UPR
	CMPB	'z'+1,R0
	BLS	UPR
	ADDB	'A'-'a',R0
UPR:
	RET

RDHEX:
	XORB	R2,R2		; Count
	XORD	R3,R3		; Value
RH0:
	MOVB	(R1),R0
	BSR	UPPER
	CMPB	'0',R0
	BHI	RHE
	CMPB	'9'+1,R0
	BHI	RH1
	CMPB	'A',R0
	BHI	RHE
	CMPB	'F'+1,R0
	BLS	RHE
	SUBB	'A'-'9'-1,R0
RH1:
	SUBB	'0',R0
	LSHD	4,R3
	ORB	R0,R3
	ADDQD	1,R1
	ADDQB	1,R2
	BR	RH0
RHE:
	RET

;;; Data area

OPNMSG:	DB	CR,LF,"Universal Monitor NS32000",CR,LF,0

PROMPT:	DB	"] ",0

IHEMSG:	DB	"Error ihex",CR,LF,0
SHEMSG:	DB	"Error srec",CR,LF,0
ERRMSG:	DB	"Error",CR,LF,0

DSEP0:	DB	" :",0
DSEP1:	DB	" : ",0

;;; Module Descriptor

MODDSC:
	DD	WORK_B		; Static base
	DD	00000000H	; Link base
	DD	CSTART		; Program base
	DD	00000000H

	IF USE_DEV_EMILY
	INCLUDE	"dev/dev_emily.asm"
	ENDIF

;;;
;;; RAM area
;;;

	ORG	WORK_B

INBUF:	DS	BUFLEN
DSADDR:	DS	4
DEADDR:	DS	4
SADDR:	DS	4
GADDR:	DS	4
HEXMOD:	DS	1

